/***************************************************************************
 *   Copyright (C) 2010 by Thomas Richard <thomas.richard@proan.be>        *
 *   Copyright (C) 2009-2010 by Peter Penz <peter.penz@gmx.at>             *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA            *
 ***************************************************************************/

#include <kaction.h>
#include <kdemacros.h>
#include <kfileitem.h>
#include <klocalizedstring.h>
#include <QDir>
#include <QString>
#include <QStringList>
#include <QLocalSocket>

#include <KDE/KPluginFactory>
#include <KDE/KPluginLoader>

#include "dolphin-box-plugin.h"

K_PLUGIN_FACTORY(DolphinBoxPluginFactory, registerPlugin<DolphinBoxPlugin>();)
K_EXPORT_PLUGIN(DolphinBoxPluginFactory("dolphin-box-plugin"))

DolphinBoxPlugin::DolphinBoxPlugin(QObject* parent, const QList<QVariant>& args) :
    KVersionControlPlugin(parent),
    m_pendingOperation(false),
    m_versionInfoHash(),
    m_browseAction(0),
    m_revisionsAction(0),
    m_showDeletedAction(0),
    m_shareAction(0),
    m_copyPublicURLAction(0),
    m_contextUrl(),
    m_socket()
{
    Q_UNUSED(args);

    m_browseAction = new KAction(this);
    m_browseAction->setIcon(KIcon("internet-web-browser"));
    m_browseAction->setText(i18nc("@item:inmenu", "Browse on Dropbox Website..."));
    connect(m_browseAction, SIGNAL(triggered()),
            this, SLOT(browseAction()));

    m_revisionsAction = new KAction(this);
    m_revisionsAction->setIcon(KIcon("view-history"));
    m_revisionsAction->setText(i18nc("@item:inmenu", "View Previous Versions..."));
    connect(m_revisionsAction, SIGNAL(triggered()),
            this, SLOT(revisionsAction()));

    m_showDeletedAction = new KAction(this);
    m_showDeletedAction->setIcon(KIcon("edit-undo"));
    m_showDeletedAction->setText(i18nc("@item:inmenu", "Show Deleted Files..."));
    connect(m_showDeletedAction, SIGNAL(triggered()),
            this, SLOT(showDeletedAction()));

    m_shareAction = new KAction(this);
    m_shareAction->setIcon(KIcon("user-group-new"));
    m_shareAction->setText(i18nc("@item:inmenu", "Share This Folder..."));
    connect(m_shareAction, SIGNAL(triggered()),
            this, SLOT(shareAction()));

    m_copyPublicURLAction = new KAction(this);
    m_copyPublicURLAction->setIcon(KIcon("klipper"));
    m_copyPublicURLAction->setText(i18nc("@item:inmenu", "Copy Public Link"));
    connect(m_copyPublicURLAction, SIGNAL(triggered()),
            this, SLOT(copyPublicURLAction()));

    m_copyGalleryURLAction = new KAction(this);
    m_copyGalleryURLAction->setIcon(KIcon("klipper"));
    m_copyGalleryURLAction->setText(i18nc("@item:inmenu", "Copy Public Gallery Link"));
    connect(m_copyGalleryURLAction, SIGNAL(triggered()),
            this, SLOT(copyGalleryURLAction()));

    m_socketpath = QDir::home().path();
    m_socketpath.append(QDir::separator()).append(".dropbox/command_socket");
    m_socketpath = QDir::toNativeSeparators(m_socketpath);
    m_socket.connectToServer(m_socketpath);

    QString dbpath = QDir::home().path();
    dbpath.append(QDir::separator()).append(".dropbox/dropbox.db");
    dbpath = QDir::toNativeSeparators(dbpath);

    m_dbfile = QFileInfo(dbpath);
    m_lastUpdate = m_dbfile.lastModified();

    //This timer defines the poll interval of the dropboxd
    m_timer = new QTimer(this);
    m_timer->start(1000);
    connect(m_timer, SIGNAL(timeout()), this, SLOT(checkIfUpdateNeeded()));
}

DolphinBoxPlugin::~DolphinBoxPlugin()
{
}

QString DolphinBoxPlugin::fileName() const
{
    return QLatin1String(".dropbox");
}

bool DolphinBoxPlugin::beginRetrieval(const QString& directory)
{
    Q_ASSERT(directory.endsWith(QLatin1Char('/')));

    //Make socket for this worker thread, this must be a new one because
    //the other one is created in the gui thread
    QLocalSocket socket;

    //Clear hasmap
    m_versionInfoHash.clear();

    //Get directory listing
    QDir dir(directory);
    QStringList files = dir.entryList();

    //Start at 2 because we're not interested in . and ..
    for(int i=2;i<files.size();++i)
    {
        QString filename = dir.absolutePath() + QDir::separator() + files.at(i);
        QString command  = "icon_overlay_file_status\npath\t" + filename;

        QString reply = sendCommand(command, &socket, false);

        //The connection failed but we still return true to avoide anoying
        // "Failed to update version information"
        if(reply.isNull())
            return true;

        QStringList parametersSplitted = reply.split('\t');
        QString state = parametersSplitted.at(1);

        VersionState versionstate = UnversionedVersion;

        if(state == "up to date")
        {
            versionstate = NormalVersion;
        }
        else if(state == "syncing")
        {
            versionstate = UpdateRequiredVersion;
        }
        else if(state == "unsyncable")
        {
            versionstate = ConflictingVersion;
        }

        m_versionInfoHash.insert(filename, versionstate);
    }

    return true;
}

void DolphinBoxPlugin::endRetrieval()
{
}

KVersionControlPlugin::VersionState DolphinBoxPlugin::versionState(const KFileItem& item)
{
    const QString itemUrl = item.localPath();
    if (m_versionInfoHash.contains(itemUrl))
    {
        return m_versionInfoHash.value(itemUrl);
    }

    return UnversionedVersion;
}

QList<QAction*> DolphinBoxPlugin::contextMenuActions(const KFileItemList& items)
{
    Q_ASSERT(!items.isEmpty());

    if(items.size() > 1 || items.size() == 0)
    {
        QList<QAction*> emptyactions;
        m_contextUrl = QString();
        return emptyactions;
    }

    KFileItem item = items.at(0);
    return getActions(item.url().path());
}

QList<QAction*> DolphinBoxPlugin::contextMenuActions(const QString& directory)
{
    return getActions(directory);
}

void DolphinBoxPlugin::browseAction()
{
    QString command  = "icon_overlay_context_action\nverb\tbrowse\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::revisionsAction()
{
    QString command  = "icon_overlay_context_action\nverb\trevisions\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::showDeletedAction()
{
    QString command  = "icon_overlay_context_action\nverb\tdeleted\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::shareAction()
{
    QString command  = "icon_overlay_context_action\nverb\tshare\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::copyPublicURLAction()
{
    QString command  = "icon_overlay_context_action\nverb\tcopypublic\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::copyGalleryURLAction()
{
    QString command  = "icon_overlay_context_action\nverb\tcopygallery\npaths\t" + m_contextUrl;
    sendCommand(command, &m_socket);
}

void DolphinBoxPlugin::checkIfUpdateNeeded()
{
    m_dbfile.refresh();
    if(m_dbfile.lastModified() != m_lastUpdate)
    {
        m_lastUpdate = m_dbfile.lastModified();
        emit versionStatesChanged();
    }
}

//Re-entrant
QString DolphinBoxPlugin::sendCommand(QString command, QLocalSocket *socket, bool inGuiThread)
{
    //If we're in the GuiThread we can't afford to wait long
    //in the worker thread we have bigger margins
    int waitTime = 500;
    if(inGuiThread)
        waitTime = 100;

    if(!socket->isOpen())
    {
        socket->connectToServer(m_socketpath);
        if(!socket->waitForConnected(waitTime))
            return QString();
    }

    socket->write(command.toUtf8());
    socket->write(QString("\n").toUtf8());
    socket->write(QString("done\n").toUtf8());
    socket->flush();

    bool ready = false;
    QString reply;

    while(!ready)
    {
        if(!socket->waitForReadyRead(waitTime))
        {
            //If we have to wait this long, the socket probably isn't open anymore (dropbox died or closed)
            socket->close();
            return QString();
        }

        reply.append(socket->readAll());

        if(reply.endsWith("done\n"))
        {
            ready = true;
        }
    }

    //Strip out \ndone\n and ok\n
    reply = reply.remove("\ndone\n");
    reply = reply.remove("ok\n");

    return reply;
}

QList<QAction*> DolphinBoxPlugin::getActions(QString url)
{
    QList<QAction*> actions;

    m_contextUrl = url;
    QString command  = "icon_overlay_context_options\npaths\t" + url;
    QString reply = sendCommand(command, &m_socket);

    QStringList splitted = reply.split('\t');

    //Skip the options keyword
    for(int i=1; i<splitted.size(); i++)
    {
        QString option = splitted.at(i);

        QStringList splittedOption = option.split('~');

        QString action;

        if(splittedOption.size() == 3)
            action = splittedOption.at(2);
        else
            continue;

        if(action == "browse")
            actions.append(m_browseAction);
        else if(action == "revisions")
            actions.append(m_revisionsAction);
        else if(action == "deleted")
            actions.append(m_showDeletedAction);
        else if(action == "share")
            actions.append(m_shareAction);
        else if(action == "copypublic")
            actions.append(m_copyPublicURLAction);
        else if(action == "copygallery")
            actions.append(m_copyGalleryURLAction);
    }

    return actions;
}
